import click
import json
import sys
import time
import requests

from .utils import print_key_value, print_status_header, print_status_row
from .network import DidimoAuth, http_get, http_post, http_post_withphoto
from .config import Config
from .helpers import get_didimo_status, download_didimo, URL, download_asset
from ._version import __version__

pass_api = click.make_pass_decorator(Config)

CONTEXT_SETTINGS = dict(help_option_names=['--help', '-h'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("-c", "--config", help="Use this configuration instead of the default one.")
@click.pass_context
def cli(ctx, config):
    """
    Create, list and download didimos
    """
    ctx.ensure_object(Config)
    if ctx.invoked_subcommand not in ["init", "version"]:
        ctx.obj.load()
        if config != None:
            ctx.obj.load_configuration(config)
        else:
            ctx.obj.load_configuration(ctx.obj.configuration)


@cli.command(short_help="Initializes configuration")
@click.argument("name", required=True)
@click.option('--host', type=URL(), prompt=True, help="API host with protocol.", default="https://api.didimo.co", show_default=True)
@click.option('--api-key', prompt="API Key", help="API Key from your credentials.")
@click.option('--api-secret', prompt="API Secret", help="API Secret from your credentials.")
@pass_api
def init(config, name, host, api_key, api_secret):
    """
    Initializes configuration

    <NAME> is the name of the configuration that will be added.
    """
    config.init(name, host, api_key, api_secret)


@cli.command()
@click.option("-n", "--number", required=False, default=1, show_default=True,
              help="Number of pages to query from the API. Each page has 10 didimos.")
@click.option("-r", "--raw", required=False, is_flag=True, default=False,
              help="Do not format output, print raw JSON response from API, ignoring --number.")
@pass_api
def list(config, number, raw):
    """
    List didimos
    """
    api_path = "/v3/didimos/"
    url = config.api_host + api_path
    r = http_get(url, auth=DidimoAuth(config, api_path))
    if raw:
        click.echo(r.text)
    else:

        if number < 1:
            sys.exit(0)

        didimos = []
        page = 1
        didimos += r.json()['didimos']

        while page != number:

            next_page = r.json()['__links']['next']
            if next_page != None:
                api_path = next_page
                url = api_path
                r = http_get(url, auth=DidimoAuth(config, api_path))
                didimos += r.json()['didimos']
                page += 1
            else:
                break

        print_status_header()
        for didimo in didimos:
            print_status_row(didimo)


@cli.command()
@click.option("-r", "--raw", required=False, is_flag=True, default=False,
              help="Do not format output, print raw JSON response from API.")
@pass_api
def account(config, raw):
    """
    Get account information
    """
    api_path = "/v3/accounts/default/status"
    url = config.api_host + api_path

    r = http_get(url, auth=DidimoAuth(config, api_path))
    response = r.json()

    if raw:
        click.echo(r.text)
    else:

        #tier = response["tier"]["name"]
        #print (response["owner_profile_uuid"])
        #print (tier)

        api_path2 = "/v3/didimos/"
        url2 = config.api_host + api_path2

        #print (url2)

        r2 = http_get(url2, auth=DidimoAuth(config, api_path2))
        didimos = r2.json()

        print_key_value("Tier", response["tier"]["name"])
        print_key_value("Points", response["balance"])
        print_key_value("Total didimos in account", didimos['total_size'])
        expiry_message = "\n(!) %s points will expire at %s" % \
            (response["next_expiration_points"],
             response["next_expiration_date"])
        click.secho(expiry_message, fg="yellow", err=True)


@cli.command(short_help="Create a didimo")
@click.argument("type", type=click.Choice(["photo"]), required=True, metavar="TYPE")
@click.argument("input", type=click.Path(exists=True), required=True)
@click.option('--depth', '-d',
              type=click.Path(), required=False,
              help="Create didimo with depth")
@click.option('--feature', '-f', multiple=True,
              type=click.Choice(
                  ["oculus_lipsync", "simple_poses", "arkit", "aws_polly"]),
              help="Create didimo with optional features. This flag can be used multiple times.")
@click.option('--max-texture', '-m', multiple=False,
              type=click.Choice(
                  ["512", "1024", "2048"]),
              help="Create didimo with optional max texture dimension. ")
@click.option('--no-download', '-n', is_flag=True, default=False,
              help="Do not download didimo")
@click.option('--no-wait', '-w', is_flag=True, default=False,
              help="Do not wait for didimo creation and do not download")
@click.option("--output", "-o", type=click.Path(), required=False,
              help="Path to download the didimo. If multiple package types "
              "are present or if the flags --no-wait or --no-download "
              "are present, this option is ignored. [default: <ID>.zip]")
@click.option('--package-type', '-p', multiple=True,
              type=click.Choice(["fbx", "gltf"]),
              help="Specify output types for this didimo. This flag can be used multiple times.", show_default=True)
@click.option("--version", "-v",
              type=click.Choice(["2.5"]),
              default="2.5",
              help="Version of the didimo.", show_default=True)
@pass_api
def new(config, type, input, depth, feature, max_texture, no_download, no_wait, output, package_type, version):
    """
    Create a didimo

    TYPE is the type of input used to create the didimo. Accepted values are:

    \b
        - photo (input must be a .jpg/.jpeg/.png)
        - depth (input must be a .png)

        For more information on the input types, visit
        https://developer.didimo.co/docs/cli\b

    INPUT is the path to the input file.

    \b
    Examples:
        Create a didimo from a photo
        $ didimo new photo /path/input.jpg

    """

    api_path = "/v3/didimos"
    url = config.api_host + api_path

    payload = {
        'input_type': 'photo'
    }

    if depth != None:
        payload["input_type"] = "rgbd"
    else:
        payload["input_type"] = "photo"

    if len(package_type) > 0:
        payload["transfer_formats"] = package_type
        package_type = package_type[0]
    else:
        package_type = "gltf"

    if max_texture != None:
        payload["max_texture_dimension"] = max_texture

    for feature_item in feature:
        payload[feature_item] = 'true'

    r = http_post_withphoto(url, config.access_key, payload, input, depth)

    didimo_id = r.json()['key']

    click.echo(didimo_id)
    if not no_wait:
        with click.progressbar(length=100, label='Creating didimo') as bar:
            last_value = 0
            while True:
                response = get_didimo_status(config, didimo_id)
                percent = response.get('percent', 100)
                update = percent - last_value
                last_value = percent
                bar.update(update)
                if response['status_message'] != "":
                    click.secho(err=True)
                    click.secho('Error: %s' %
                                response["status_message"], err=True, fg='red')
                    sys.exit(1)
                if response['status'] == 'done':
                    break
                time.sleep(2)
        if not no_download:
            if output == None:
                output = "%s_%s.zip" % (didimo_id, package_type)
            download_didimo(config, didimo_id, "", output)


@cli.command(short_help='Get status of didimos')
@click.argument("id", required=True, nargs=-1)
@click.option("-r", "--raw", required=False, is_flag=True, default=False,
              help="Do not format output, print raw JSON response from API.")
@click.option("-s", "--silent", required=False, is_flag=True, default=False,
              help="Do not print anything. See help text for exit codes.")
@pass_api
def status(config, id, raw, silent):
    """
    Get status of didimos

    <ID> is the didimo ID to get information.

    Multiple didimo IDs are accepted, separated by a space or newline

    If <ID> is the character "-", read the IDs from STDIN.

    When using the --silent flag, the following exit code rules are applied:

    \b
      - 0: Every didimo is in "Done" state
      - 1: At least one didimo is in "Error" state
      - 2: At least one didimo is in "Pending" or "Processing" state
    """
    didimos = []

    ids = set(id)

    # read didimo ids if used with a pipe
    if "-" in ids:
        ids = sys.stdin.readlines()

    for didimo in ids:

        response = get_didimo_status(config, didimo.rstrip())

        # TODO
        # Remove this block when /status endpoint is consistent with /list
        if response['status_message'] != "":
            response["key"] = didimo.rstrip()
            response["status"] = "error"

        didimos.append(response)
        if silent:
            if response["status"] == "error":
                click.echo("Error on \"%s\"" % response["key"], err=True)
                sys.exit(1)
            if response["status"] in ["pending", "processing"]:
                sys.exit(2)

    if silent:
        sys.exit(0)

    if raw:
        click.echo(json.dumps(didimos, indent=4))
    else:
        print_status_header()
        for didimo in didimos:
            print_status_row(didimo)


@cli.command(short_help="Get or set configuration")
@click.argument("name", required=False)
@pass_api
def config(config, name):
    """
    Get or set configuration

    With no arguments, list all available configurations.

    If <NAME> is provided, set that as the default configuration.
    """
    if name == None:
        config.list_configurations()
    else:
        click.echo("Setting default configuration to \"%s\"" % name, err=True)
        config.save_configuration(name)
        click.secho("Configuration file saved.", fg='blue', err=True)


@cli.command()
@click.argument("id", required=True)
@click.option("-o", "--output", type=click.Path(),
              help="Download path. [default: <ID>.zip]")
@click.option('--package-type', '-p',
              type=click.Choice(["fbx", "gltf"]),
              help="Specify output type for this didimo.", show_default=True)
@pass_api
def download(config, id, output, package_type):
    """
    Download a didimo

    <ID> is the didimo ID.

    When given the "-" character, read the didimo
    ID from STDIN.
    """

    if id == "-":
        id = sys.stdin.readlines()[0].rstrip()

    if output is None:
        output = "%s_%s.zip" % (id, package_type)
    download_didimo(config, id, package_type, output)


@cli.command()
@pass_api
def version(config):
    """
    Print version and exit
    """
    print(__version__)
    sys.exit(0)


@cli.group()
@pass_api
def execute(config):
    """
    Execute on-demand features on didimos
    """
    pass


@execute.command(short_help="Produce high fidelity hairs on a didimo")
@click.argument("input", type=click.Path(exists=True), required=True)
@pass_api
def hairsdeform(config, input):
    """
    Produce high fidelity hairsdeform on a didimo

    <INPUT> is your deformation file.

    Returns the didimo asset ID that you can use with other commands and hairsdeform package.
    """

    api_path = "/v3/assets"
    url = config.api_host + api_path

    payload = {
        'input_type': 'hairs_deform'
    }

    files = [('template_deformation', (input, open(
        input, 'rb'), 'application/octet-stream'))]

    headers = {
        'DIDIMO-API-KEY': config.access_key
    }

    r = requests.request("POST", url, headers=headers,
                         data=payload, files=files)

    if r.status_code != 201:
        click.secho('Error %d' % r.status_code, err=True, fg='red')
        click.echo(r.text)
        sys.exit(1)

    response = r.json()

    key = response['key']
    url = ""
    for package_itm in r.json()['transfer_formats']:
        url = package_itm["__links"]["self"]
        break

    click.echo(response['key'])
    output = "%s.zip" % key

    click.echo("Creating package file.")
    time.sleep(15)
    download_asset(config, url, api_path, output)


@execute.command(short_help="Deform a model to match a didimo shape")
@click.argument("vertex", required=True, type=click.Path(exists=True))
@click.argument("user_asset", required=True, type=click.Path(exists=True))
@pass_api
def vertexdeform(config, vertex, user_asset):
    """
    Deform a model to match a didimo shape

    <VERTEX> is your vertex file.
    <USER_ASSET> is your asset file.

    Returns an asset ID of the deformed vertex that you can use with other commands and the package.
    """

    api_path = "/v3/assets"
    url = config.api_host + api_path

    payload = {'input_type': 'vertex_deform'}

    files = [
        ('template_deformation', (vertex, open(
            vertex, 'rb'), 'application/octet-stream')),
        ('user_asset', (user_asset, open(user_asset, 'rb'), 'application/octet-stream'))
    ]

    headers = {
        'DIDIMO-API-KEY': config.access_key
    }

    r = requests.request("POST", url, headers=headers,
                         data=payload, files=files)

    if r.status_code != 201:
        click.secho('Error %d' % r.status_code, err=True, fg='red')
        click.echo(r.text)
        sys.exit(1)

    response = r.json()

    key = response['key']
    url = ""
    for package_itm in r.json()['transfer_formats']:
        url = package_itm["__links"]["self"]
        break

    click.echo(response['key'])
    output = "%s.zip" % key

    click.echo("Creating package file.")
    time.sleep(15)
    download_asset(config, url, api_path, output)



import click
import sys

from .network import DidimoAuth, http_get
from urllib import parse as urlparse


class URL(click.ParamType):
    name = "url"

    def convert(self, value, param, ctx):
        if not isinstance(value, tuple):
            value_parsed = urlparse.urlparse(value)
            if value_parsed.scheme not in ("http", "https"):
                self.fail("invalid URL scheme", param, ctx)
            if value_parsed.netloc == "":
                self.fail("host is empty", param, ctx)
        return value


def get_didimo_status(config, id):
    api_path = "/v3/didimos/" + id
    url = config.api_host + api_path
    r = http_get(url, auth=DidimoAuth(config, api_path))
    return r.json()


def download_didimo(config, id, package_type, output_path):
    api_path = "/v3/didimos/" + id
    url = config.api_host + api_path
    r = http_get(url, auth=DidimoAuth(config, api_path))


    for package_itm in r.json()['transfer_formats']:
        s3url = ""
        output_path = id+"_"+package_itm["name"] + ".zip"

        if package_type == None or len(package_type) == 0:
            s3url = package_itm["__links"]["self"]
        else:
            if package_itm["name"] == package_type:
                s3url = package_itm["__links"]["self"]

        if s3url != "":
            print ("downloading....")
            with http_get(s3url, auth=DidimoAuth(config, api_path)) as r:
                r.raise_for_status()
                zipsize = int(r.headers.get('content-length', 0))
                with click.open_file(output_path, 'wb') as f:
                    label = "Downloading %s" % id
                    with click.progressbar(length=zipsize, label=label) as bar:
                        for chunk in r.iter_content(chunk_size=2048):
                            size = f.write(chunk)
                            bar.update(size)
            click.secho('Downloaded to %s' % output_path, fg='blue', err=True)

        # else:
            # print ("Unable to download")


def download_asset(config, asset_url, api_path, output_path):

    if asset_url != "":
        print ("downloading....")
        with http_get(asset_url, auth=DidimoAuth(config, api_path)) as r:
            r.raise_for_status()
            zipsize = int(r.headers.get('content-length', 0))
            with click.open_file(output_path, 'wb') as f:
                label = "Downloading %s" % id
                with click.progressbar(length=zipsize, label=label) as bar:
                    for chunk in r.iter_content(chunk_size=2048):
                        size = f.write(chunk)
                        bar.update(size)
        click.secho('Downloaded to %s' % output_path, fg='blue', err=True)

    else:
        print ("Unable to download")



import click
import requests
import hmac
import time
import sys
import platform
from hashlib import sha256

from ._version import __version__


class DidimoAuth(requests.auth.AuthBase):
    def __init__(self, config, path):
        self.config = config
        self.path = path

    def __call__(self, r):
        access_key = self.config.access_key
        r.headers["didimo-api-key"] = access_key
        r.headers["User-Agent"] = "didimo-cli/%s (%s, %s)" % (__version__,
                                                              platform.python_version(),
                                                              platform.system())

        return r


def http_get(url, **kwargs):
    try:
        r = requests.get(url, **kwargs)
        if r.status_code == 200:
            return r
        else:
            click.secho('Error %d' % r.status_code, err=True, fg='red')
            click.echo(r.text)
            sys.exit(1)
    except:
        click.echo("A Network Error Has Occured")
        sys.exit(1)


def http_post(url, **kwargs):
    r = requests.post(url, **kwargs)
    if r.status_code == 200:
        return r
    else:
        click.secho('Error %d' % r.status_code, err=True, fg='red')
        click.echo(r.text)
        sys.exit(1)


def http_post_withphoto(url, access_key, payload, photo, photo_depth):

    if photo_depth != None:
        files = [
            ('photo', (photo, open(photo, 'rb'), 'image/jpeg')),
            ('depth', (photo_depth, open(photo_depth, 'rb'), 'image/png'))
        ]
    else:
        files = [('photo', (photo, open(photo, 'rb'), 'image/jpeg'))]

    headers = {
        'DIDIMO-API-KEY': access_key
    }

    r = requests.request("POST", url, headers=headers,
                         data=payload, files=files)

    if r.status_code == 201:
        return r
    else:
        click.secho('Error %d' % r.status_code, err=True, fg='red')
        click.echo("An error has occured. Please check your API key")
        sys.exit(1)



import click
from datetime import datetime


def print_key_value(key, value, fg="white"):
    click.secho("%s: " % key, bold=True, nl=False, fg=fg, err=True)
    click.secho(str(value), err=True)


def print_status_header():
    click.secho("{:<18} │ {:<5} │ {:<7} │ {:^4} │ {:^4}  "
                .format(
                    'didimo ID', 'Type', 'Percent', 'Status', 'Created At'),
                bold=True, err=True)
    click.secho("─" * 112)


def print_status_row(didimo):

    # Remove when /status endpoint is consistent with /list
    if didimo.get('status') == "error":
        click.secho("{:<18} │ {:<5} │ {:<7} │ {:<6} │ {:^4}  "
                    .format(didimo['key'], "-", "-",
                             didimo['status'].title(), "-"), fg="red")
        return

    color = 'white'
    if didimo['status'] == "error":
        color = 'red'
    elif didimo['status'] == "done":
        color = 'green'
    elif didimo['status'] == "processing":
        color = 'yellow'

    didimo_type = didimo['input_type']
    if didimo['input_type'] == "lofimesh_texture":
        didimo_type = 'sony'
    elif didimo['input_type'] == "hifimesh_texture_photo":
        didimo_type = 'scan'
    elif didimo['input_type'] == "blendshapes":
        didimo_type = 'blend'
    elif didimo['input_type'] == "vertexdeform":
        didimo_type = 'vertx'

    #created_at = datetime.strptime(didimo['created_at'], "%Y-%m-%d %H:%M:%S")
    click.secho("{:<18} │ {:<5} │ {:<7} │ {:^7}│ {:^7}"
                .format(
                    didimo['key'], didimo_type.title(),
                    didimo['percent'],
                    didimo['status'].title(),
                    didimo['created_at']),
                fg=color)



from setuptools import setup
import pathlib

pkg_vars = {}
here = pathlib.Path(__file__).parent.resolve()

with open(here / 'cli/_version.py', encoding='utf-8') as f:
    exec(f.read(), pkg_vars)

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name="didimo-cli",
    version=pkg_vars["__version__"],
    description="Command-line interface to interact with Didimo API",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/didimoinc/cli",
    project_urls={
        'Bug Reports': 'https://github.com/didimoinc/cli/issues',
        'Source': 'https://github.com/didimoinc/cli',
    },
    author="Didimo, Inc.",
    author_email="support@didimo.co",
    keywords='cli, didimo, api',
    license='Apache 2.0',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Games/Entertainment',
        'Topic :: Multimedia :: Graphics',
        'Programming Language :: Python :: 3'
    ],
    packages=["cli"],
    include_package_data=True,
    install_requires=[
      "click",
      "requests"
    ],
    entry_points="""
        [console_scripts]
        didimo=cli.cli:cli
    """
)



import click
import json
import sys
import os

from pathlib import Path


class Config(object):
    def __init__(self):
        self.access_key = ""
        self.secret_key = ""
        self.configuration = ""
        self.api_host = ""

    def init(self, configuration, host, api_key, api_secret):
        config_dir = Path.home() / ".didimo"
        config_dir.mkdir(parents=True, exist_ok=True)

        config_file = config_dir / ("%s.json" % configuration)
        if config_file.exists():
            overwrite = click.confirm(
                'Found a configuration for %s. Overwrite?' % configuration)
            if not overwrite:
                click.echo('Discarding changes', err=True)
                sys.exit(0)

        with click.open_file(config_file, "w") as f:
            config = {"host": host, "access_key": api_key,
                      "secret_key": api_secret}
            json.dump(config, f, indent=2)
            click.secho("Wrote configuration at %s" % config_file, err=True)

        cli_config_file = config_dir / "cli.json"
        with click.open_file(cli_config_file, "w") as f:
            config = {"default": configuration}
            json.dump(config, f, indent=2, sort_keys=True)
            click.secho("Set \"%s\" as default configuration" %
                        configuration, err=True)

    def load(self):
        try:
            with click.open_file(Path.home() / ".didimo" / "cli.json", "r") as f:
                config = json.load(f)
                self.configuration = config["default"]
        except FileNotFoundError:
            click.secho(
                "CLI configuration file not found. Run `didimo init`", err=True, fg='red')
            sys.exit(1)
        except json.decoder.JSONDecodeError:
            click.secho("Error decoding JSON from \"cli.json\"",
                        err=True, fg='red')
            sys.exit(1)

    def list_configurations(self):
        config_dir = Path.home() / ".didimo"
        envs = sorted([f for f in os.listdir(
            config_dir) if f.endswith(".json")])
        for e in envs:
            env = e.split('/')[-1].split('.')[0]
            if env != "cli":
                if env == self.configuration:
                    click.secho(env, bold=True, err=True, nl=False)
                    click.secho(" [x]", bold=True, fg='blue')
                else:
                    click.secho(env, err=True)

    def load_configuration(self, configuration):
        config_file = Path.home() / ".didimo" / (configuration + ".json")
        if not config_file.exists():
            click.secho("No configuration file for \"%s\"." %
                        configuration, err=True, fg='red')
            sys.exit(1)

        try:
            with click.open_file(config_file, 'r') as f:
                config = json.load(f)
                self.api_host = config.get("host", "")
                self.access_key = config.get("access_key", "")
                self.secret_key = config.get("secret_key", "")
                if self.access_key == "" or self.secret_key == "":
                    click.secho("No access key or secret key for \"%s\" configuration" %
                                configuration, err=True, fg='red')
                    sys.exit(1)
                if self.api_host == "":
                    click.secho("No API host for \"%s\" configuration" %
                                configuration, err=True, fg='red')
                    sys.exit(1)
        except json.decoder.JSONDecodeError:
            click.secho("Error decoding configuration file %s" %
                        config_file, err=True, fg='red')
            sys.exit(1)

    def save_configuration(self, configuration):
        config_file = Path.home() / ".didimo" / (configuration + ".json")
        if not config_file.exists():
            click.secho("No configuration file for \"%s\"." %
                        configuration, err=True, fg='red')
            sys.exit(1)

        cli_file = Path.home() / ".didimo" / "cli.json"
        with click.open_file(cli_file, "w") as f:
            config = {"default": configuration}
            json.dump(config, f, indent=2, sort_keys=True)
